#ifndef SHAPES_H
#define SHAPES_H

//#define USE_SIMD

#include <cmath>
#ifdef USE_SIMD
    #include <xmmintrin.h>
#endif

struct Vec3
{
    inline Vec3() : x( 0.0f ), y( 0.0f ), z( 0.0f ), w( 0.0f ) {}

    inline Vec3( float ax, float ay, float az ) : x( ax ), y( ay ), z( az ) {}

    inline Vec3 operator*( const Vec3& v ) const
    {
        return Vec3( x * v.x, y * v.y, z * v.z );
    }

    inline Vec3 operator*( float f ) const
    {
        return Vec3( x * f, y * f, z * f );
    }

    inline Vec3 operator/( float f ) const
    {
        return Vec3( x / f, y / f, z / f );
    }

    inline Vec3 operator-() const
    {
        return Vec3( -x, -y, -z );
    }

    inline Vec3 operator-( const Vec3& other ) const
    {
        return Vec3( x - other.x, y - other.y, z - other.z );
    }

    inline void operator+=( const Vec3& other )
    {
        x += other.x;
        y += other.y;
        z += other.z;
    }

    inline void operator-=( const Vec3& other )
    {
        x -= other.x;
        y -= other.y;
        z -= other.z;
    }

#ifdef USE_SIMD
    inline Vec3 operator+( const Vec3& aOther ) const
    {
        __m128 self   = _mm_load_ps( &x );
        __m128 other  = _mm_load_ps( &aOther.x );
        __m128 result = _mm_add_ps( self, other );

        Vec3 out;
        _mm_store_ps( &out.x, result );
        return out;
    }
#else
    inline Vec3 operator+( const Vec3& other ) const
    {
        return Vec3( x + other.x, y + other.y, z + other.z );
    }
#endif

    Vec3 cross( const Vec3& b ) const
    {
        Vec3 out;
        
    	out.x = y * b.z - z * b.y;
        out.y = z * b.x - x * b.z;
        out.z = x * b.y - y * b.x;
        
        return out;
    }

    inline float length() const
    {
        return sqrt( x * x + y * y + z * z );
    }

#ifdef USE_SIMD
    float dot( const Vec3& aOther ) const
    {
        __m128 self   = _mm_load_ps( &x );
        __m128 other  = _mm_load_ps( &aOther.x );
        __m128 result = _mm_add_ps( self, other );
    return 1.0f;
    }
#else
    float dot( const Vec3& other ) const
    {
        return x * other.x + y * other.y + z * other.z;
    }
#endif

    void clampToColor()
    {
        if (x < 0.0f)
        {
            x = 0.0f;
        }

        if (y < 0.0f)
        {
            y = 0.0f;
        }
        
        if (z < 0.0f)
        {
            z = 0.0f;
        }

        if (x > 1.0f)
        {
            x = 1.0f;
        }

        if (y > 1.0f)
        {
            y = 1.0f;
        }

        if (z > 1.0f)
        {
            z = 1.0f;
        }
    }

    void normalize()
    {
        const float iLength = 1.0f / sqrt( x * x + y * y + z * z );
        x *= iLength;
        y *= iLength;
        z *= iLength;
    }

    Vec3 reflect( const Vec3& normal ) const
    {
        return *this - normal * normal.dot( *this ) * 2.0f;
    }
#ifdef USE_SIMD
    __declspec( align( 16 ) ) float x;
#else
    float x;
#endif
    float y, z, w;
};

enum class Type
{
    Plane,
    Sphere,
    Triangle
};

struct Shape
{
    Shape() :
        emission( Vec3( 0.0f, 0.0f, 0.0f ) )
      , radius( 0.0f )
      , type( Type::Sphere )
    {}

    static Shape createSphere( const Vec3& aPosition, float aRadius, const Vec3& aColor, const Vec3& aEmission )
    {
        Shape shape;
        shape.type = Type::Sphere;
        shape.position = aPosition;
        shape.radius = aRadius;
        shape.color = aColor;
        shape.emission = aEmission;
        return shape;
    }
    

    static Shape createPlane( const Vec3& aPosition, const Vec3& aNormal, const Vec3& aColor, const Vec3& aEmission )
    {
        Shape shape;
        shape.type = Type::Plane;
        shape.position = aPosition;
        shape.normal = aNormal;
        shape.color = aColor;
        shape.emission = aEmission;
        return shape;
    }

    static Shape createTriangle( const Vec3& v0, const Vec3& v1, const Vec3& v2, const Vec3& aColor, const Vec3& aEmission )
    {
        Shape shape;
        shape.type = Type::Triangle;
        shape.emission = aEmission;
        shape.color = aColor;
        shape.v0 = v0;
        shape.v1 = v1;
        shape.v2 = v2;
        
        Vec3 va = v0;
        Vec3 vb = v1;
        Vec3 vc = v2;
        
        vb -= va;
        vc -= va;
        
        shape.normal = vb.cross( vc );
        
        shape.normal.normalize();

        return shape;
    }

    float intersect( const Vec3& rayOrigin, const Vec3& rayDir, bool primaryRay ) const;

    void getCollisionInfo( const Vec3& rayOrigin, const Vec3& rayDir, float distance, Vec3& outHitpoint, Vec3& outNormal ) const;

    Vec3 position;
    Vec3 color;
    Vec3 normal;
    Vec3 emission;
    Vec3 v0;
    Vec3 v1;
    Vec3 v2;
    float radius;
    Type type;

    // First-ray optimization.
    float eyeDotNormal;
};

#endif
