#include "Tracer.h"
#include <cstdlib>
#include <cmath>
#include <vector>

std::vector< Shape > shapes;

void initShapes()
{
    shapes.push_back( Shape::createSphere( Vec3( -10.0f, 0.0f,  0.0f ), 15.0f, Vec3( 0.0f, 0.0f, 1.0f ), Vec3( 0.0f, 0.0f, 0.0f ) ) );
    shapes.push_back( Shape::createSphere( Vec3(   0.0f, -0.5f, -10.0f ), 15.0f, Vec3( 0.0f, 1.0f, 0.0f ), Vec3( 0.0f, 0.0f, 0.0f ) ) );
    shapes.push_back( Shape::createSphere( Vec3(  10.0f, 0.0f,  0.0f ), 15.0f, Vec3( 1.0f, 0.0f, 0.0f ), Vec3( 0.0f, 0.0f, 0.0f ) ) );
    
    shapes.push_back( Shape::createPlane( Vec3( 0.0f, 5.0f, 0.0f ), Vec3( 0.0f, -1.0f, 0.0f ), Vec3( 0.5f, 0.5f, 0.5f ), Vec3( 0.0f, 0.0f, 0.0f ) ) );
    shapes.push_back( Shape::createPlane( Vec3( 0.0f, -5.0f, 0.0f ), Vec3( 0.0f, 1.0f, 0.0f ), Vec3( 0.5f, 0.5f, 0.5f ), Vec3( 1.0f, 1.0f, 1.0f ) ) ); // Ceiling.
    shapes.push_back( Shape::createPlane( Vec3( -15.0f, 0.0f, 0.0f ), Vec3( 1.0f, 0.0f, 0.0f ), Vec3( 1.0f, 0.0f, 0.0f ), Vec3( 0.0f, 0.0f, 0.0f ) ) );
    shapes.push_back( Shape::createPlane( Vec3( 15.0f, 0.0f, 0.0f ), Vec3( -1.0f, 0.0f, 0.0f ), Vec3( 0.0f, 1.0f, 0.0f ), Vec3( 0.0f, 0.0f, 0.0f ) ) );
    shapes.push_back( Shape::createPlane( Vec3( 0.0f, 0.0f, 10.0f ), Vec3( 0.0f, 0.0f, -1.0f ), Vec3( 1.0f, 1.0f, 1.0f ), Vec3( 0.0f, 0.0f, 0.0f ) ) );
    
    // Top.
    shapes.push_back( Shape::createTriangle( Vec3( -5.0f, 3.0f, -5.0f ), Vec3( 5.0f, 3.0f, -20.0f ), Vec3( 5.0f, 3.0f, -5.0f ),
                                             Vec3( 0.0f, 0.0f, 1.0f ), Vec3( 0.0f, 0.0f, 0.0f ) ) );
    shapes.push_back( Shape::createTriangle( Vec3( -5.0f, 3.0f, -5.0f ), Vec3( -5.0f, 3.0f, -20.0f ), Vec3( 5.0f, 3.0f, -20.0f ),
                                             Vec3( 0.0f, 0.0f, 1.0f ), Vec3( 0.0f, 0.0f, 0.0f ) ) );
    // Front.
    shapes.push_back( Shape::createTriangle( Vec3( -5.0f, 3.0f, -20.0f ), Vec3( 5.0f, 5.0f, -20.0f ), Vec3( 5.0f, 3.0f, -20.0f ),
                                             Vec3( 1.0f, 0.0f, 0.0f ), Vec3( 0.0f, 0.0f, 0.0f ) ) );
    shapes.push_back( Shape::createTriangle( Vec3( -5.0f, 3.0f, -20.0f ), Vec3( -5.0f, 5.0f, -20.0f ), Vec3( 5.0f, 5.0f, -20.0f ),
                                             Vec3( 1.0f, 0.0f, 0.0f ), Vec3( 0.0f, 0.0f, 0.0f ) ) );
    
	srand( 1337 );
}

float randomFloat( float aMin, float aMax )
{
	const float r = (float)rand() / (float)RAND_MAX;
	return aMin + r * (aMax - aMin);
}


Vec3 randomRayInHemisphere( const Vec3& aNormal )
{
    Vec3 v2( randomFloat( -1.0f, 1.0f ), randomFloat( -1.0f, 1.0f ), randomFloat( -1.0f, 1.0f ) );
    v2.normalize();

    while (v2.dot( v2 ) > 1.0f)
    {
        v2 = Vec3( randomFloat( -1.0f, 1.0f ), randomFloat( -1.0f, 1.0f ), randomFloat( -1.0f, 1.0f ) );
        v2.normalize();
    }

    return v2 * (v2.dot( aNormal ) < 0.0f ? -1.0f : 1.0f);
}


const Shape* findClosestShape( const Vec3& aOrigin, const Vec3& aDirection, float& aOutClosestDistance, bool aPrimaryRay )
{
    aOutClosestDistance = 1000000.0f;
    const Shape* closestShape = nullptr;

    for (auto& shape : shapes)
    {
        const float distance = shape.intersect( aOrigin, aDirection, aPrimaryRay );
        if (distance > 0.0029f && distance < aOutClosestDistance)
        {
            aOutClosestDistance = distance;
            closestShape = &shape;
        }
    }

    return closestShape;
}


void precalcFrame( const Vec3& aRayOrigin )
{
    for (auto& shape : shapes)
    {
        if (shape.type == Type::Plane)
        {
            shape.eyeDotNormal = (shape.position - aRayOrigin).dot( shape.normal );
        }
    }
}


Vec3 pathTraceRay( const Vec3& aOrigin, const Vec3& aDirection, int aReflectionCount )
{
    if (aReflectionCount > 10)
    {
        return Vec3( 0.0f, 0.0f, 0.0f );
    }

    float distance;
    const Shape* closestShape = findClosestShape( aOrigin, aDirection, distance, aReflectionCount == 0 );
    
    if (!closestShape)
    {
        return Vec3( 0.0f, 0.0f, 0.0f );
    }

    Vec3 hitPoint, hitNormal;
    closestShape->getCollisionInfo( aOrigin, aDirection, distance, hitPoint, hitNormal );

    const Vec3 newDirection = randomRayInHemisphere( hitNormal );

    const float lambert = 1.0f; //newDirection.dot( hitNormal );

    if (closestShape->emission.x > 0.0f || closestShape->emission.y > 0.0f || closestShape->emission.z > 0.0f)
    {
        return closestShape->emission;
    }

    return closestShape->color * lambert * pathTraceRay( hitPoint, newDirection, aReflectionCount + 1 ) + closestShape->emission;
}
