//
// Path Tracer
//
// \author Timo Wiren
// \date 2012-12-22
// Licence: zlib Licence

#include <cassert>
#include <algorithm>
#include <chrono>
#include <iostream>
#include <thread>
#include <vector>
#include <SDL/SDL.h>
#include "Shapes.h"
#include "Tracer.h"

#define WIDTH 640
#define HEIGHT 480

Uint32* pixels = NULL;
int gSamples = 1;
Vec3 gRayOrigin( 0.0f, 0.0f, -32.0f );
std::vector< Vec3 > gAccumColor1( WIDTH * HEIGHT );

void initShapes();

unsigned int encodeColor( const Vec3& aColor )
{
#ifdef __APPLE__
    return (((unsigned int)(aColor.z * 255.0f)) << 8) | (((unsigned int)(aColor.y * 255.0f)) << 16) | (((unsigned int)(aColor.x * 255.0f)) << 24);
#else
    return ((unsigned int)(aColor.z * 255.0f)) | (((unsigned int)(aColor.y * 255.0f)) << 8) | (((unsigned int)(aColor.x * 255.0f)) << 16);    
#endif
}


void render( int aBeginY, int aEndY )
{
    for (int y = aBeginY; y < aEndY; ++y)
    {
        for (int x = 0; x < WIDTH; ++x)
        {
            Vec3 rayDir( (float)x - WIDTH / 2.0f, (float)y - HEIGHT / 2.0f, 512.0f );
            rayDir.normalize();

            Vec3 color = pathTraceRay( gRayOrigin, rayDir, 0 );
            color.clampToColor();

            gAccumColor1[ y * WIDTH + x ] += color;
        }
    }
}


void updatePixels()
{
    for (int y = 0; y < HEIGHT; ++y)
    {
        const int yw = y * WIDTH;

        for (int x = 0; x < WIDTH; ++x)
        {
            pixels[ yw + x ] = encodeColor( gAccumColor1[ yw + x ] / (float)gSamples );
        }
    }
}


int main( int argc, char* argv[] )
{
    SDL_Surface* screen = SDL_SetVideoMode( WIDTH, HEIGHT, 32, SDL_DOUBLEBUF | SDL_HWSURFACE | SDL_ASYNCBLIT );
    assert( screen );
    pixels = (Uint32*)screen->pixels;
    assert( pixels );

    initShapes();

    SDL_Event event = { 0 };

    std::fill( std::begin( gAccumColor1 ), std::end( gAccumColor1 ), Vec3() );

    while (event.type != SDL_KEYDOWN)
    {
        SDL_PollEvent( &event );
     
        ++gSamples;

        precalcFrame( gRayOrigin );

        auto beginRender = std::chrono::system_clock::now();
        
        // For some reason, threaded rendering produces errorneous output on
        // Visual Studio 2012 Update 1, but works on other compilers.
#ifndef _WIN32
        std::thread thread1( render, 0, HEIGHT / 2 );
        std::thread thread2( render, HEIGHT / 2, HEIGHT );
        thread1.join();
        thread2.join();
#else
        render( 0, HEIGHT );
#endif
        auto duration = std::chrono::system_clock::now() - beginRender;
        auto millis = std::chrono::duration_cast< std::chrono::milliseconds >( duration ).count();
        std::cout << "Tracing time: " << millis << " ms." << std::endl;
        updatePixels();

        SDL_Flip( screen );
    }
    
    SDL_Quit();
    return 0;
}

